import typing
from typing import Any, get_args, get_origin, TypeVar, Iterable, TYPE_CHECKING, Type

from pydantic import BaseModel

from pydantic._internal._model_construction import ModelMetaclass
from pydantic.fields import FieldInfo
from pydantic_core import PydanticUndefined
from xsentinels import Default
from xsentinels.default import DefaultType

from .config import PartialConfigDict
from .sentinels import Missing, MissingType, AutoPartialExcludeMarker

from logging import getLogger

if TYPE_CHECKING:
    from .partial import PartialModel


# metaclass to make all fields in a model optional, useful for PATCH requests
class PartialMeta(ModelMetaclass):
    """ Metaclass of `pydantic_partials.partial.PartialModel`, used to support partial fields,
        including the ability ot automatically apply `pydantic_partials.partial.Partial`[...] to fields
        via `pydantic_partials.config.PartialConfigDict.auto_partials` configuration option.
    """
    model_partial_fields: typing.ClassVar[set[str]]
    """ Set of strings representing field names that can be missing from validation/serialization.
        If they are missing, they will return the `Missing` sentinel value and the field will be entirely
        skipped when serializing the model.

        If not missing, they will return their normal value, be included in serializing the model
        and otherwise act normally.

        If any type has their type-hint unioned with `MissingType` it will be a partial field.
        If `auto_add_missing_type` is True (default is True), then all fields on model will automatically
        be unioned with MissingType.
    """

    config_dict: typing.ClassVar[PartialConfigDict]
    """ Adding some extra/new config options on-top of the already pre-existing Pydantic ones.
        Seemed like a good place to put them, as it deals with configuration of the class model behavior.
    """

    def __new__(
            cls,
            name: str,
            bases: tuple[type],
            namespaces: dict[str, Any],
            *,

            auto_partials: bool | DefaultType = Default,
            auto_partials_exclude: Iterable[str] | DefaultType = Default,

            # A private/internal detail for generic base subclasses that want to also change the fields,
            # this prevents having to rebuild the class a second time; if this is True then the subclass
            # is responsible for calling rebuild when finished.
            ___PartialMeta__delay_rebuild: bool = False,

            **kwargs
    ):
        """

        Args:
            auto_partials: For more details see `pydantic_partials.config.PartialConfigDict.auto_partials`.
                If `Default`: Inherit behavior from parent/model_config; otherwise defaults to `False`.
                If `True`: Will automatically make all fields on the model `Partial`.
                If `False` (default): User needs to mark individual fields as `Partial` where they want.

            auto_partials_exclude: If `auto_partials` is True, or you inherit from `AutoPartialModel`
                (which simply has `auto_partials` set to `True` by default for you) then this field will be consulted.

                If auto partials is `False` (default) then `auto_partials_exclude` is ignored,
                but we will still keep track of any provided `auto_partials_exclude` in case there is a
                subclass that does enable auto_partials.

                This is a set of strings of field names to exclude from automatic partials.
                If you explicitly mark a field as a Partial, this won't affect that. THis only effects
                automatically applied partial fields.

                You can also use `pydantic_partials.partial.AutoPartialExclude` to more easily mark fields as excluded.
                For more details see `pydantic_partials.config.PartialConfigDict.auto_partials_exclude`.

            **kwargs: Passed along other class arguments to Pydantic and any __init_subclass__ methods.
        """
        # Create the class first...
        cls: 'Type[PartialModel]' = super().__new__(cls, name, bases, namespaces, **kwargs)  # type: ignore

        model_config: PartialConfigDict = cls.model_config  # type: ignore

        # We now have the fields Pydantic found and can now easily add our MissingType's as needed.
        # Then a model_rebuild is forced to update the class schema to include the MissingType/LazyType annotations,
        # and any needed changes to the field default value.

        if auto_partials is not Default:
            model_config['auto_partials'] = bool(auto_partials)

        if auto_partials_exclude:
            model_config['auto_partials_exclude'] = set(auto_partials_exclude)  # type: ignore

        # Inherit any pre-existing `auto_partials_exclude` items.
        # For now if someone wants to override this, they can simply manually do this on subclass:
        #     `field_name: Partial[...]`
        # I might consider a simple way to inform us to not inherit excludes, but for now keeping things simple.
        final_auto_exclude = model_config.get('auto_partials_exclude', set())
        for c in cls.__mro__:  # type: ignore
            parent_config: PartialConfigDict | None = getattr(c, 'model_config', None)
            if not parent_config:
                continue

            final_auto_exclude.update(parent_config.get('auto_partials_exclude', set()))

        need_rebuild = False

        partial_fields = set()
        for k, v in cls.model_fields.items():  # type: ignore
            field_type = v.annotation
            origin = get_origin(field_type)
            if origin is None:
                if field_type is MissingType:
                    partial_fields.add(k)

            # TODO: Check that `field_type` is a union?
            for arg_type in get_args(field_type):
                if arg_type is MissingType:
                    partial_fields.add(k)

            for mdv in v.metadata:
                # If we find the marker AND we are not already marked as a partial_field,
                # add field to auto-exclude list.
                if mdv is AutoPartialExcludeMarker and k not in partial_fields:
                    final_auto_exclude.add(k)

        if final_auto_exclude:
            model_config['auto_partials_exclude'] = final_auto_exclude

        # Default to `False` for auto-partials.
        final_partial_auto = model_config.get('auto_partials', False)
        if final_partial_auto is not False:
            # I'll be putting in more options for `final_partial_auto` in the near future,
            # so just check for `is True` for now.
            if final_partial_auto is not True:
                raise ValueError(f'Invalid/Unknown `partial_auto` config value ({final_partial_auto}), use bool value.')

            for k, v in cls.model_fields.items():  # type: ignore
                if k in partial_fields:
                    # The field is already a Partial
                    continue

                if k in final_auto_exclude:
                    # The field is excluded.
                    continue

                if v.default is PydanticUndefined and v.default_factory is None:
                    v.annotation = v.annotation | MissingType
                    partial_fields.add(k)
                    need_rebuild = True

        fields: dict[str, FieldInfo] = cls.model_fields  # type: ignore
        for k in partial_fields:
            v = fields[k]
            if v.default is PydanticUndefined and v.default_factory is None:
                v.default = Missing
                need_rebuild = True

        cls.model_partial_fields = partial_fields

        if need_rebuild and not ___PartialMeta__delay_rebuild:
            cls.model_rebuild(force=True)  # type: ignore
        return cls




import typing
from typing import Any, TypeVar, Annotated, TypeAlias

from pydantic import BaseModel

from .config import PartialConfigDict
from .meta import PartialMeta
from .sentinels import Missing, MissingType, AutoPartialExcludeMarker

from logging import getLogger

log = getLogger(__name__)


PM = TypeVar('PM')
Partial: TypeAlias = PM | MissingType
""" Can be used to manually mark a variable as Partial, which means it can have a `Missing`
    assigned to it. 
"""

PME = TypeVar('PME')
AutoPartialExclude = Annotated[PME, AutoPartialExcludeMarker]


class PartialModel(
    BaseModel,

    # Need metaclass to examine fields for missing type
    # and also to auto-add missing type if desired.
    metaclass=PartialMeta,

    # Needed so `Missing` default values will be validated and therefore the `PydanticOmit` will be
    # raised and inform Pydantic to ignore/omit the field value.
    validate_default=True
):
    """
    Class Args:

    - auto_partials: For more details see `pydantic_partials.config.PartialConfigDict.auto_partials`.
        - If `Default`: Inherit behavior from parent/model_config; otherwise defaults to `True`.
        - If `True` (default): Will automatically make all fields on the model `Partial`.
        - If `False`: User needs to mark individual fields as `Partial` where they want.
    """

    config_dict: typing.ClassVar[PartialConfigDict]

    def __init__(self, *args, **kwargs):
        """ Pydantic partial model class, with ability to easily dynamically omit fields when serializing a model.
        """
        super().__init__(*args, **kwargs)

    if not typing.TYPE_CHECKING:

        # We put `__getattr__` in a non-TYPE_CHECKING block because otherwise, mypy allows arbitrary attribute access
        def __getattr__(self, item: str) -> Any:
            try:
                return super().__getattr__(item)
            except AttributeError:
                # Determine if value is 'Missing' or if we need to propagate the error...
                if item in type(self).model_partial_fields:
                    return Missing
                raise

        def __setattr__(self, key, value):
            # Right now, if a `raise PydanticOmit()` is done during validation while `validate_assignment=True` is on,
            # it will raise an error in Pydantic. I am assuming this is a bug,
            # because validation via `Model.model_validate()` works just fine when raising a `PydanticOmit`.
            # For now, as a work-around, I look for `Missing` values when set on object and ignore the set if found.
            #
            # Whenever a 'Missing' value is set, the user wants to ignore the field value, so we delete it if it exists.
            # I might rather keep the `Missing' value on the object, but there is currently
            # no easy way to exclude the value from the json output until they get something like this done:
            # https://github.com/pydantic/pydantic/discussions/5461#discussioncomment-7411977
            #
            # But on the other hand, I can use `def __getattr__` to lazily resolve any Missing values
            # instead of using __getattribute__; which is much more efficient... So I may stick with it in any case.
            # Lazily resolving would be for something like an ORM like object that wants to lazily lookup relationships.
            if value is Missing and key in type(self).model_partial_fields:
                try:
                    delattr(self, key)
                except AttributeError:
                    pass
            else:
                super().__setattr__(key, value)


class AutoPartialModel(PartialModel, auto_partials=True):
    pass



from typing import Any, Type, TypeVar, Annotated

from pydantic import GetCoreSchemaHandler
from pydantic_core import core_schema, PydanticOmit
from xsentinels.sentinel import Sentinel


class MissingType(Sentinel):
    """ Class/Type of `Missing`, a sentinel that is used to indicate if a field is missing its value
        in a `pydantic_partials.partial.PartialModel` subclass.
    """
    # Notes/See (https://docs.pydantic.dev/latest/concepts/json_schema/#modifying-the-schema).

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Type[Any], handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        assert source_type is MissingType
        # We never want to serialize any Missing values.
        return core_schema.with_info_after_validator_function(
            cls._validate,
            core_schema.is_instance_schema(cls=MissingType),  # core_schema.any_schema(),
            # serialization=core_schema.plain_serializer_function_ser_schema(
            #     cls._serialize,
            #     # return_schema=core_schema.AnySchema
            # )
        )

    @staticmethod
    def _validate(value: Any, info: core_schema.ValidationInfo) -> 'MissingType':
        if value is Missing:
            # Keeps the associated attribute 'deleted/omitted' from model.
            raise PydanticOmit()

        # If we somehow get a non-Missing value (should not happen), return it unchanged.
        return value

    @staticmethod
    def _serialize(value: Any) -> 'MissingType':
        # Keeps the associated attribute 'deleted/omitted' from model.
        # raise PydanticOmit()

        # Return same value we got, when requested to try and serialize real data it's not 'Missing',
        # and if it is, that means `Value` is `Missing` and we are fine to return that also.
        # I would love to rase a `raise PydanticOmit()` if `value` is `Missing`, but that does not currently work
        # for Pydantic, so I just return the `Missing` unchanged for now.
        # If there is a serialization error later on about how Pydantic can't turn `Missing` into a `Json`
        # then we can debug the situation
        # (right now PartialModel should delete any attributes that are set to `Missing` to work around this limitation)
        return value


Missing = MissingType()
""" Returned as attribute value when attribute value is missing. Can also be set on attribute to indicate it's missing.
"""


class AutoPartialExcludeMarkerType(Sentinel):
    pass


AutoPartialExcludeMarker = AutoPartialExcludeMarkerType()
""" Used by `pydantic_partials.partial.AutoPartialExclude` to mark a field as excluded from automatic partials.
    See `pydantic_partials.config.PartialConfigDict.auto_partials_exclude` for more details.
"""



import typing

from pydantic import ConfigDict
from xsentinels.default import DefaultType, Default


class PartialConfigDict(ConfigDict, total=False):
    auto_partials: bool
    """
    Defaults to `True`.
    
    - If `True` (default): When class is created, all required attributes will be made `Partial`,
        and their default set to `Missing`. This means that all attributes that don't have any
        default or default_factory defined will automatically have `MissingType`/`LazyType` added
        to their annotation and field default set to `Missing`/`Lazy`.
        
        Default value will only be modified if it was not already set to something by the user.
        It needs to be modified if nothing is defined to ensure the field is not required.
        This allows Pydantic to create an instance of the model without a value defined
        for partial fields.
    
    - If `False`: Class won't automatically add `Partial` fields.
        Instead adding them to individual fields will be up to the user.
        
        The default value of any fields set by user to be `Partial` will still be adjusted
        unless the user already a default value defined by user for field.
        This ensures that Pydantic won't require the field be assigned a value.
    """

    auto_partials_exclude: set[str]
    """ A set of field names to exclude from the `auto_partials` functionality.
        
        All parent classes `partials_exclude` string sets will be combined
        together for a final exclusion list.
        
        Any partials set manually/directly via the fields annotation will still be a partial field
        regardless if it's referenced in this `auto_partials_exclude` set or not. This only effects
        the way automatic partials are applied.
        
        You can also use `pydantic_partials.partial.AutoPartialExclude` to more easily mark fields as excluded.
    """
